#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
import datetime


class BaseType:
    UTC = datetime.timezone.utc
    EP_YEAR = 1970
    EP_MONTH = 1
    EP_DAY = 1
    EP = datetime.datetime(EP_YEAR, EP_MONTH, EP_DAY, 0, 0, 0, tzinfo=UTC)
